/**
 * @(#) ExampleVariationPointsProvider.java  2015/11/23
 * Copyright (c) 2015 No Magic, Inc. All Rights Reserved.
 */
package com.nomagic.magicdraw.variants.examples;

import com.nomagic.magicdraw.core.Project;
import com.nomagic.magicdraw.properties.PropertyManager;
import com.nomagic.magicdraw.uml.ElementFinder;
import com.nomagic.magicdraw.variants.PropertyNames;
import com.nomagic.magicdraw.variants.evaluation.BooleanEvaluator;
import com.nomagic.magicdraw.variants.evaluation.Evaluator;
import com.nomagic.magicdraw.variants.variationpoints.VariationPoint;
import com.nomagic.magicdraw.variants.variationpoints.VariationPointsFactory;
import com.nomagic.magicdraw.variants.variationpoints.VariationPointsProvider;
import com.nomagic.uml2.ext.jmi.helpers.ModelHelper;
import com.nomagic.uml2.ext.jmi.helpers.StereotypesHelper;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.Class;
import com.nomagic.uml2.ext.magicdraw.classes.mdkernel.*;
import com.nomagic.uml2.ext.magicdraw.mdprofiles.Stereotype;

import javax.annotation.CheckForNull;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

/**
 * An example variation points provider implementation.
 * It returns all elements in a given scope as variable elements.
 * variation points are created only for specific named elements for sake of demonstration.
 *
 * @author Tomas Lukosius
 */
public class ExampleVariationPointsProvider implements VariationPointsProvider
{
	private VariationPointsFactory factory = new VariationPointsFactory();

	@Override
	public String getName()
	{
		return "Example Variation Points Provider";
	}

	@Override
	public void setUp(Project project, @CheckForNull PropertyManager propertyManager)
	{
	 /*
		 called just before the start of the transformation. This method is called automatically, when transforming from MagicDraw Model Transformation dialog.
		 if transformation is triggered not from MagicDraw Model Transformation dialog, this method should be called before transformation.
		 Initialize resources that are required for getVariabilityElements() and getVariationPoints() methods
	 */
	}

	@Override
	public void tearDown(Collection<VariationPoint> failed)
	{
		// automatically called after transformation is complete. Resources should be released here
	}

	/**
	 * Collect elements in scope, that will be transformed
	 *
	 * @param scope where to collect variable elements
	 * @return collection of variable elements
	 */
	@Override
	public Collection<Element> getVariabilityElements(Collection<Element> scope)
	{
		// define, what elements will be transformed in a given scope

		// it's recommended to return only those elements, that have to be transformed.
		// What is more, single element should be included once.
		// Then transformation will execute faster.
		Collection<Element> variabilityElements = new HashSet<>();
		for (Element scopeElement : scope)
		{
			Collection<? extends Element> elements = ModelHelper.getElementsOfType(scopeElement, null, true, true);
			variabilityElements.addAll(elements);
		}
		return variabilityElements;
	}

	/**
	 * Collect and return all variation points for element
	 *
	 * @param element element, for which variation points are collected
	 * @return collection of variation points
	 */
	@Override
	public Collection<VariationPoint> getVariationPoints(Element element)
	{
		// here we create variation points for given element
		if (element instanceof NamedElement)
		{
			final String name = ((NamedElement) element).getName();
			// com.nomagic.magicdraw.variants.PropertyNames contains constants of properties, supported in transformations
			switch (name)
			{
				case "existence":
					return Collections.<VariationPoint>singletonList(
							factory.createExistenceVariationPoint(element, createExistenceChangeEvaluator()));
				case "name":
					return Collections.<VariationPoint>singletonList(
							factory.createPropertyVariationPoint(element, PropertyNames.NAME, createNameChangeEvaluator()));
				case "default value":
					return Collections.<VariationPoint>singletonList(
							factory.createDefaultValueVariationPoint(element, createDefaultValueChangeEvaluator()));
				case "multiplicity":
					return Collections.<VariationPoint>singletonList(
							factory.createMultiplicityVariationPoint(element, createMultiplicityChangeEvaluator()));
				case "changeType":
					return Collections.<VariationPoint>singletonList(
							factory.createTypeVariationPoint(element, createTypeChangeEvaluator()));
				case "removeFlow":
					return Collections.<VariationPoint>singletonList(
							factory.createActivityBranchExistenceVariationPoint(element, createExistenceChangeEvaluator()));
				case "changeToSubtype":
					return Collections.<VariationPoint>singletonList(
							factory.createTypeVariationPoint(element, createChangeTypeToFirstSubtypeEvaluator()));
				case "stereotypedClass":
					return Arrays.<VariationPoint>asList(
							factory.createTagVariationPoint(element, getAppliedStereotype(element), "someElementValue",
															createElementValueChangeEvaluator()),
							factory.createTagVariationPoint(element, getAppliedStereotype(element), "someCount",
															createIntegerIncrementEvaluator())
					);
			}
		}
		return Collections.emptyList();
	}

	/**
	 * Return first stereotype applied to given element. Element may have multiple stereotypes, but we applied only one to keep example
	 * as simple as possible.
	 *
	 * @param element given element with one stereotype applied
	 * @return first applied stereotype
	 */
	private Stereotype getAppliedStereotype(Element element)
	{
		return StereotypesHelper.getStereotypes(element).iterator().next();
	}

	/*
		below are some examples of an evaluator creation.
		An evaluator returns value, which will be applied to given element during transformation.
		An evaluator receives a current value parameter, which can be used to calculate a value, that is applied during transformation.
		com.nomagic.magicdraw.variants.variationpoints.VariationPointsFactory has methods to create all kinds of currently supported variation points.
	 */
	private BooleanEvaluator createExistenceChangeEvaluator()
	{
		return new BooleanEvaluator()
		{
			@Override
			public Boolean evaluate(Element element)
			{
				return false;
			}
		};
	}

	private Evaluator<Object> createNameChangeEvaluator()
	{
		return new Evaluator<Object>()
		{
			@Override
			public Object evaluate(Element element, Object currentValue)
			{
				return currentValue == null ? "" : currentValue.toString() + " changed";
			}
		};
	}

	private Evaluator<Object> createDefaultValueChangeEvaluator()
	{
		return new Evaluator<Object>()
		{
			@Override
			public Object evaluate(Element element, Object currentValue)
			{
				return 10;
			}
		};
	}

	private Evaluator<String> createMultiplicityChangeEvaluator()
	{
		return new Evaluator<String>()
		{
			@Override
			public String evaluate(Element element, Object currentValue)
			{
				return "2";
			}
		};
	}

	private Evaluator<Type> createTypeChangeEvaluator()
	{
		return new Evaluator<Type>()
		{
			@Override
			public Type evaluate(Element element, Object currentValue)
			{
				return (Type) ElementFinder.find(Project.getProject(element).getModel(), Class.class, "TypeB", true);
			}
		};
	}

	private Evaluator<Type> createChangeTypeToFirstSubtypeEvaluator()
	{
		return new Evaluator<Type>()
		{
			@Override
			public Type evaluate(Element element, @CheckForNull Object currentValue)
			{
				if (currentValue instanceof Classifier)
				{
					Collection<Classifier> derivedClassifiers = ModelHelper.getDerivedClassifiers((Classifier) currentValue);
					if (!derivedClassifiers.isEmpty())
					{
						return derivedClassifiers.iterator().next();
					}
				}
				return (Type) currentValue;
			}
		};
	}

	private Evaluator<Object> createElementValueChangeEvaluator()
	{
		return new Evaluator<Object>()
		{
			@Override
			public Object evaluate(Element element, @CheckForNull Object currentValue)
			{
				return ElementFinder.find(Project.getProject(element).getModel(), Class.class, "TypeB", true);
			}
		};
	}

	private Evaluator<Object> createIntegerIncrementEvaluator()
	{
		return new Evaluator<Object>()
		{
			@Override
			public Object evaluate(Element element, @CheckForNull Object currentValue)
			{
				if (currentValue instanceof Integer)
				{
					return (Integer) currentValue + 1;
				}
				return 0;
			}
		};
	}
}
